/***************************************************************************
 * Copyright 2012 by
 *  + Christian-Albrechts-University of Kiel
 *    + Department of Computer Science
 *      + Software Engineering Group
 *  and others.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ***************************************************************************/

package kieker.webgui.common;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.math.RoundingMode;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.concurrent.ConcurrentHashMap;

import kieker.analysis.AnalysisController;
import kieker.analysis.model.analysisMetaModel.MIAnalysisMetaModelFactory;
import kieker.analysis.model.analysisMetaModel.MIDependency;
import kieker.analysis.model.analysisMetaModel.MIProject;
import kieker.analysis.model.analysisMetaModel.impl.MAnalysisMetaModelFactory;
import kieker.common.logging.Log;
import kieker.common.logging.LogFactory;
import kieker.webgui.common.exception.LibraryAlreadyExistingException;
import kieker.webgui.common.exception.NewerProjectException;
import kieker.webgui.common.exception.ProjectAlreadyExistingException;

import org.primefaces.model.UploadedFile;

/**
 * This is a singleton class for the access to the file system. It makes sure that the necessary directories for the execution of the application exist. <b>Do
 * not</b> remove directories created from this manager during runtime! Directories are created during first access to the class.<br>
 * This class uses also a fine grained synchronization to handle the access to the projects.
 * 
 * @author Nils Christian Ehmke
 */
public final class FSManager {

	/**
	 * This is the extension of the KAX-files.
	 */
	private static final String KAX_EXTENSION = "kax";
	/**
	 * This is the extension of the library-files.
	 */
	private static final String LIB_EXTENSION = "jar";
	/**
	 * This is the name of the directory for the libraries.
	 */
	private static final String LIB_DIRECTORY = "lib";
	/**
	 * This is the name of the root-directory.
	 */
	private static final String ROOT_DIRECTORY = "data";
	/**
	 * This is the name of the directory for the views.
	 */
	private static final String VIEW_DIRECTORY = "view";
	/**
	 * This is the buffer (in bytes) used to copy files.
	 */
	private static final int BUF_SIZE_BYTES = 1024;
	/**
	 * This is the factory we will use to create the elements for the projects.
	 */
	private final MIAnalysisMetaModelFactory factory = new MAnalysisMetaModelFactory();
	/**
	 * This map contains one lock-object for every project (by name). This works at every project name may exist only once.
	 */
	private final ConcurrentHashMap<String, Object> projectLocksMap = new ConcurrentHashMap<String, Object>();
	/**
	 * This is the log object used to log messages, warnings etc.
	 */
	private static final Log LOG = LogFactory.getLog(FSManager.class);
	/**
	 * This is the singleton instance of this class.
	 */
	private static final FSManager INSTANCE = new FSManager();

	/**
	 * Creates a new instance of this class.
	 */
	private FSManager() {
		// Check for our root-directory and create it if necessary
		final File rootDir = new File(FSManager.ROOT_DIRECTORY);
		if (!rootDir.exists()) {
			final boolean result = rootDir.mkdir();
			if (!result) {
				FSManager.LOG.error("Could not create root directory.");
			}
		}
	}

	/**
	 * Delivers the singleton instance of this class.
	 * 
	 * @return The only instance of {@link FSManager}.
	 */
	public static FSManager getInstance() {
		return FSManager.INSTANCE;
	}

	/**
	 * This method adds a new project to the application. If the method fails due to an {@link IOException}, it will make sure that the project-directories will be
	 * removed as if the method would never have been called.
	 * 
	 * @param project
	 *            The project to be added.
	 * @throws IOException
	 *             If a write-error occurred during the creation.
	 * @throws ProjectAlreadyExistingException
	 *             If a project with the same name exists already.
	 */
	public void addProject(final String project) throws IOException, ProjectAlreadyExistingException {
		// Get the lock for the new project
		final Object lock = this.getLock(project);

		// Assemble the paths to the directory and to the files for the given project
		final File projectDir = new File(FSManager.ROOT_DIRECTORY + File.separator + project);
		final File projectFile = new File(FSManager.ROOT_DIRECTORY + File.separator + project + File.separator + project + "." + FSManager.KAX_EXTENSION);
		final File libDir = new File(FSManager.ROOT_DIRECTORY + File.separator + project + File.separator + FSManager.LIB_DIRECTORY);
		final File viewDir = new File(FSManager.ROOT_DIRECTORY + File.separator + project + File.separator + FSManager.VIEW_DIRECTORY);

		// We need an "empty" project in order to save it.

		final MIProject emptyProject = this.factory.createProject();

		// Create the directories and the empty file atomically
		synchronized (lock) {
			// Make sure that the project doesn't exist already
			if (projectDir.exists()) {
				throw new ProjectAlreadyExistingException("The project with the name '" + project + "' exists already.");
			}

			// Create the directories
			final boolean projDirCreated = projectDir.mkdir();
			final boolean libDirCreated = libDir.mkdir();
			final boolean viewDirCreated = viewDir.mkdir();
			// The following part is only necessary to calm FindBugs...
			@SuppressWarnings("unused")
			final boolean createResults = projDirCreated && libDirCreated && viewDirCreated;

			// Now the empty project file
			try {
				AnalysisController.saveToFile(projectFile, emptyProject);
			} catch (final IOException ex) {
				// Something went wrong. Remove the directories and files!
				final boolean viewDirDeleted = viewDir.delete();
				final boolean libDirDeleted = libDir.delete();
				// Keep in mind that the potential remains of the file have to be deleted before the directory.
				final boolean projectFileDeleted = projectFile.delete();
				final boolean projectDeleted = projectDir.delete();

				// The following part is only necessary to calm FindBugs...
				@SuppressWarnings("unused")
				final boolean deleteResults = viewDirDeleted && libDirDeleted && projectFileDeleted && projectDeleted;
				// Rethrow the exception in order to inform the caller of this method
				throw ex;
			}
		}
	}

	/**
	 * This method opens a project atomically, using the current version on the FS.
	 * 
	 * @param project
	 *            The project to be loaded.
	 * @return An instance of {@link MIProject} containing all informations from the project.
	 * @throws IOException
	 *             If something went wrong during opening the file.
	 */
	public MIProject openProject(final String project) throws IOException {
		// Get the lock for the given project
		final Object lock = this.getLock(project);

		// Assemble the path to the KAX-file
		final File kaxFile = new File(FSManager.ROOT_DIRECTORY + File.separator + project + File.separator + project + "." + FSManager.KAX_EXTENSION);

		// Load it atomically
		synchronized (lock) {
			return AnalysisController.loadFromFile(kaxFile);
		}
	}

	/**
	 * This method tries to save the given project.
	 * 
	 * @param projectName
	 *            The name of the project to be saved.
	 * @param project
	 *            The project to be saved.
	 * @param timeStamp
	 *            The time stamp, the given project has been opened (necessary to check for a newer version).
	 * @param overwriteNewerProject
	 *            This flag determines whether a newer project should be overwritten.
	 * @throws IOException
	 *             If something went wrong during saving the project.
	 * @throws NewerProjectException
	 *             This exception is raised if the time stamp of the project on the file system is newer than the given one <b>and</b> if the flag
	 *             {@code overwriteNewerProject} is false. If it is true, a newer version will be overwritten.
	 */
	public void saveProject(final String projectName, final MIProject project, final long timeStamp, final boolean overwriteNewerProject) throws IOException,
			NewerProjectException {
		// Get the lock
		final Object lock = this.getLock(projectName);
		synchronized (lock) {
			// Check for a newer version first
			final long currTimeStamp = this.getCurrTimeStamp(projectName);
			if (!overwriteNewerProject && (currTimeStamp > timeStamp)) {
				throw new NewerProjectException("The project with the name '" + projectName + "' has a newer version on the FS.");
			}

			// Everything seems to be okay (or we should overwrite the current file). Try to save it. Assemble the path to the KAX-file for this purpose.
			final File kaxFile = new File(FSManager.ROOT_DIRECTORY + File.separator + projectName + File.separator + projectName + "." + FSManager.KAX_EXTENSION);

			// Try to save it.
			AnalysisController.saveToFile(kaxFile, project);
		}
	}

	/**
	 * Returns a list containing all available projects on the FS as a string.
	 * 
	 * @return A list of string containing all available project.
	 */
	public Collection<? extends String> getAllProjects() {
		final List<String> result = new ArrayList<String>();

		// Get all directories within our root-dir
		final File files[] = new File(FSManager.ROOT_DIRECTORY).listFiles();
		for (final File file : files) {
			if (file.isDirectory()) {
				result.add(file.getName());
			}
		}

		return result;
	}

	/**
	 * This method returns the current time stamp of the given project. The time stamp is the last modification time of the KAX-file.
	 * 
	 * @param project
	 *            The project whose time stamp should be delivered.
	 * @return The time stamp of the last modification.
	 */
	public long getCurrTimeStamp(final String project) {
		// Assemble the path to the kax-file of the given project
		final String fileName = FSManager.ROOT_DIRECTORY + File.separator + project + File.separator + project + "." + FSManager.KAX_EXTENSION;

		// Try to find this file
		final File file = new File(fileName);

		// Get the time stamp atomically
		final Object lock = this.getLock(project);

		final long ts;
		synchronized (lock) {
			ts = file.lastModified();
		}

		// Return the time stamp
		return ts;
	}

	/**
	 * This method can be used to get the lock for a given project. If the lock doesn't exist already, a new one will be created.
	 * 
	 * @param project
	 *            The project whose lock should be delivered.
	 * @return A lock object for the project.
	 */
	private Object getLock(final String project) {
		final Object newLock = new Object();
		final Object existLock = this.projectLocksMap.putIfAbsent(project, newLock);

		return (existLock != null) ? existLock : newLock;
	}

	/**
	 * This method can be used to upload a new library.
	 * 
	 * @param srcFile
	 *            The file to be used as a new library.
	 * @param project
	 *            The name of the project for the new lib.
	 * @return An instance of {@link MIDependency} if everything went well.
	 * @throws LibraryAlreadyExistingException
	 *             If a library with the same name for the project exists already.
	 * @throws IOException
	 *             If something went wrong during writing the file.
	 */
	public MIDependency uploadLibrary(final UploadedFile srcFile, final String project) throws LibraryAlreadyExistingException, IOException {
		// Get the lock for this project
		final Object lock = this.getLock(project);

		// Prepare the files
		final File libDir = new File(FSManager.ROOT_DIRECTORY + File.separator + project + File.separator + FSManager.LIB_DIRECTORY);
		final File dstFile = new File(libDir, srcFile.getFileName());

		// Now copy the file - if it doesn't already exist
		synchronized (lock) {
			if (dstFile.exists()) {
				throw new LibraryAlreadyExistingException("The library with the name '" + srcFile.getFileName() + "' exists already.");
			}

			BufferedInputStream in = null;
			BufferedOutputStream out = null;

			try {
				// Get the streams.
				in = new BufferedInputStream(srcFile.getInputstream());
				out = new BufferedOutputStream(new FileOutputStream(dstFile));
				final byte[] buf = new byte[FSManager.BUF_SIZE_BYTES];
				int count;

				// Transfer the file.
				while ((count = in.read(buf)) != -1) {
					out.write(buf, 0, count);
				}
			} finally {
				// Try to make sure that the streams will be closed.
				if (in != null) {
					in.close();
				}
				if (out != null) {
					out.close();
				}
			}
		}

		final MIDependency dependency = this.factory.createDependency();
		dependency.setFilePath(dstFile.getName());
		return dependency;
	}

	/**
	 * This method delivers all available libraries for the given project as a pair of strings. The first element is the name of the library, the second one the size
	 * of the lib in MiBytes and as a human readable string. If the given name is null, an empty list will be returned.
	 * 
	 * @param projectName
	 *            The name of the project.
	 * @return The libraries of the given project.
	 */
	public List<Pair<String, String>> getLibraries(final String projectName) {
		final List<Pair<String, String>> result = new ArrayList<Pair<String, String>>();
		if (projectName != null) {
			// Get the lock for the project
			final Object lock = this.getLock(projectName);

			synchronized (lock) {
				// Run through the libs and put them into our list.
				final File libDir = new File(FSManager.ROOT_DIRECTORY + File.separator + projectName + File.separator + FSManager.LIB_DIRECTORY);
				final File files[] = libDir.listFiles();
				if (files != null) {
					for (final File file : files) {
						if (file.getName().endsWith("." + FSManager.LIB_EXTENSION)) {
							final float len = file.length() / 1024.0f / 1024.0f;
							final NumberFormat numberFormat = new DecimalFormat("0.00");
							numberFormat.setRoundingMode(RoundingMode.DOWN);
							final String lenStr = numberFormat.format(len);
							result.add(new Pair<String, String>(file.getName(), lenStr));
						}
					}
				}
			}
		}

		return result;
	}

	/**
	 * This method can be used to deliver the class loader for the given project. The class loader is able to load from all libraries within the given project.
	 * 
	 * @param projectName
	 *            The name of the project.
	 * @return An instance of {@link URLClassLoader} with all available libraries.
	 */
	public ClassLoader getClassLoader(final String projectName) {
		// Get the lock for the project
		final Object lock = this.getLock(projectName);
		final List<URL> libs = new ArrayList<URL>();

		// Collect all libraries of the project
		synchronized (lock) {
			// Run through the libs and put them into our list.
			final File libDir = new File(FSManager.ROOT_DIRECTORY + File.separator + projectName + File.separator + FSManager.LIB_DIRECTORY);
			final File files[] = libDir.listFiles();
			if (files != null) {
				for (final File file : files) {
					if (file.getName().endsWith("." + FSManager.LIB_EXTENSION)) {
						try {
							libs.add(file.toURL());
						} catch (final MalformedURLException ex) {
							ex.printStackTrace();
						}
					}
				}
			}

			// Now assemble the URL class loader
			return AccessController.doPrivileged(new PrivilegedAction<URLClassLoader>() {
				@Override
				public URLClassLoader run() {
					return new URLClassLoader(libs.toArray(new URL[libs.size()]), AnalysisController.class.getClassLoader());
				}
			});
		}
	}

	/**
	 * This method can be used to deliver the fully qualified URL of a given dependency for a given project.
	 * 
	 * @param lib
	 *            The library whose URL should be delivered.
	 * @param project
	 *            The corresponding project of the library.
	 * @return The URL to the given library if everything went well.
	 * @throws MalformedURLException
	 *             If the URL is for some reason invalid.
	 */
	public URL getURL(final MIDependency lib, final String project) throws MalformedURLException {
		final File file = new File(FSManager.ROOT_DIRECTORY + File.separator + project + File.separator + FSManager.LIB_DIRECTORY + File.separator
				+ lib.getFilePath());
		return file.toURL();
	}

	/**
	 * This method delivers all available libraries of the given project as a list of {@link MIDependency}.
	 * 
	 * @param projectName
	 *            The name of the project whose libraries should be delivered.
	 * @return A list with all available libraries.
	 */
	public List<MIDependency> getModelLibraries(final String projectName) {
		// Get the lock for the project
		final Object lock = this.getLock(projectName);
		final List<MIDependency> libs = new ArrayList<MIDependency>();

		// Collect all libraries of the project
		synchronized (lock) {
			// Run through the libs and put them into our list.
			final File libDir = new File(FSManager.ROOT_DIRECTORY + File.separator + projectName + File.separator + FSManager.LIB_DIRECTORY);
			final File files[] = libDir.listFiles();
			if (files != null) {
				for (final File file : files) {
					if (file.getName().endsWith("." + FSManager.LIB_EXTENSION)) {
						final MIDependency dep = this.factory.createDependency();
						dep.setFilePath(file.getName());
						libs.add(dep);
					}
				}
			}
			return libs;
		}
	}

}
